from .base import Surrogate

from .GPR import GPR, GPRSlidingWindow, GPROnlineInsert, GPROnlineStandard
from .NoGradient import NoGradient
from .NoModel import NoModel
from .RBFRegression import RBFRegression
